@@ -414,10 +414,8 @@ static int builtin_read(struct child_prog *child)
 
 	if (child->argv[1]) {
 		/* argument (VAR) given: put "VAR=" into buffer */
-		strcpy(string, child->argv[1]);
+		snprintf(string, sizeof(string)-1, "%s=", child->argv[1]);
 		len = strlen(string);
-		string[len++] = '=';
-		string[len]   = '\0';
 		fgets(&string[len], sizeof(string) - len, stdin);	/* read string */
 		newlen = strlen(string);
 		if(newlen > len)
@@ -1212,7 +1210,7 @@ static int pseudo_exec(struct child_prog *child)
 	 */
 	for (x = bltins; x->cmd; x++) {
 		if (strcmp(child->argv[0], x->cmd) == 0 ) {
-			exit(x->function(child));
+			_exit(x->function(child));
 		}
 	}
 
@@ -1220,7 +1218,7 @@ static int pseudo_exec(struct child_prog *child)
 	for (x = bltins_forking; x->cmd; x++) {
 		if (strcmp(child->argv[0], x->cmd) == 0) {
 			applet_name=x->cmd;
-			exit (x->function(child));
+			_exit (x->function(child));
 		}
 	}
 #ifdef BB_FEATURE_SH_STANDALONE_SHELL
@@ -1256,7 +1254,11 @@ static int pseudo_exec(struct child_prog *child)
 #endif
 
 	execvp(child->argv[0], child->argv);
-	perror_msg_and_die("%s", child->argv[0]);
+
+	/* Do not use perror_msg_and_die() here, since we must not 
+	 * call exit() but should call _exit() instead */
+	fprintf(stderr, "%s: %m\n", child->argv[0]);
+	_exit(EXIT_FAILURE);
 }
 
 static void insert_job(struct job *newjob, int inbg)
@@ -1462,8 +1464,11 @@ static int busy_loop(FILE * input)
 			while (!job_list.fg->progs[i].pid ||
 				   job_list.fg->progs[i].is_stopped == 1) i++;
 
-			if (waitpid(job_list.fg->progs[i].pid, &status, WUNTRACED)<0)
-				perror_msg_and_die("waitpid(%d)",job_list.fg->progs[i].pid);
+			if (waitpid(job_list.fg->progs[i].pid, &status, WUNTRACED)<0) {
+				if (errno != ECHILD) {
+					perror_msg_and_die("waitpid(%d)",job_list.fg->progs[i].pid);
+				}
+			}
 
 			if (WIFEXITED(status) || WIFSIGNALED(status)) {
 				/* the child exited */
@@ -1500,7 +1505,7 @@ static int busy_loop(FILE * input)
 	free(command);
 
 	/* return controlling TTY back to parent process group before exiting */
-	if (tcsetpgrp(shell_terminal, parent_pgrp))
+	if (tcsetpgrp(shell_terminal, parent_pgrp) && errno != ENOTTY)
 		perror_msg("tcsetpgrp");
 
 	/* return exit status if called with "-c" */
