@@ -31,7 +31,7 @@ extern int vdprintf(int d, const char *format, va_list ap)
 	char buf[BUF_SIZE];
 	int len;
 
-	len = vsprintf(buf, format, ap);
+	len = vsnprintf(buf, sizeof(buf), format, ap);
 	return write(d, buf, len);
 }
 #endif
